/*
 * (C) Copyright 2011
 * CompuLab Ltd. <www.compulab.co.il>
 * Ilya Ledvich <ilya@compulab.co.il>
 * Igor Grinberg <grinberg@compulab.co.il>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc.
 */

#include <common.h>
#include <asm/arch/bits.h>
#include <asm/arch/mux.h>
#include <asm/arch/sys_proto.h>
#include <asm/arch/sys_info.h>
#include <asm/arch/mem.h>

#include "common.h"
#include "prcm.h"
#include "memory.h"
#include "mux.h"
#include "board.h"

/*
 * DPLL(1-4) PARAM TABLES
 * Each of the tables has (ordered) M, N, FREQSEL, M2 values defined
 * for OPP (1.2V).
 * The values are defined for all possible sysclk and for ES1 and ES2.
 */
static struct dpll_param cm_t3530_mpu_dpll_param[] = {
	/* 12MHz */
	{0x0FE, 0x07, 0x05, 0x01}, /* ES1 */
	{0x0FA, 0x05, 0x07, 0x01}, /* ES2 */
	/* 13MHz */
	{0x17D, 0x0C, 0x03, 0x01}, /* ES1 */
	{0x1F4, 0x0C, 0x03, 0x01}, /* ES2 */
	/* 19.2MHz */
	{0x179, 0x12, 0x04, 0x01}, /* ES1 */
	{0x271, 0x17, 0x03, 0x01}, /* ES2 */
	/* 26MHz */
	{0x17D, 0x19, 0x03, 0x01}, /* ES1 */
	{0x0FA, 0x0C, 0x07, 0x01}, /* ES2 */
	/* 38.4MHz */
	{0x1FA, 0x32, 0x03, 0x01}, /* ES1 */
	{0x271, 0x2F, 0x03, 0x01}  /* ES2 */
};

static struct dpll_param cm_t3530_iva_dpll_param[] = {
	/* 12MHz */
	{0x07D, 0x05, 0x07, 0x01}, /* ES1 */
	{0x0B4, 0x05, 0x07, 0x01}, /* ES2 */
	/* 13MHz */
	{0x0FA, 0x0C, 0x03, 0x01}, /* ES1 */
	{0x168, 0x0C, 0x03, 0x01}, /* ES2 */
	/* 19.2MHz */
	{0x082, 0x09, 0x07, 0x01}, /* ES1 */
	{0x0E1, 0x0B, 0x06, 0x01}, /* ES2 */
	/* 26MHz */
	{0x07D, 0x0C, 0x07, 0x01}, /* ES1 */
	{0x0B4, 0x0C, 0x07, 0x01}, /* ES2 */
	/* 38.4MHz */
	{0x13F, 0x30, 0x03, 0x01}, /* ES1 */
	{0x0E1, 0x17, 0x06, 0x01}  /* ES2 */
};

static struct dpll_param cm_t3530_core_dpll_param[] = {
	/* 12MHz */
	{0x19F, 0x0E, 0x03, 0x01}, /* ES1 */
	{0x0A6, 0x05, 0x07, 0x01}, /* ES2 */
	/* 13MHz */
	{0x1B2, 0x10, 0x03, 0x01}, /* ES1 */
	{0x14C, 0x0C, 0x03, 0x01}, /* ES2 */
	/* 19.2MHz */
	{0x19F, 0x17, 0x03, 0x01}, /* ES1 */
	{0x19F, 0x17, 0x03, 0x01}, /* ES2 */
	/* 26MHz */
	{0x1B2, 0x21, 0x03, 0x01}, /* ES1 */
	{0x0A6, 0x0C, 0x07, 0x01}, /* ES2 */
	/* 38.4MHz */
	{0x19F, 0x2F, 0x03, 0x01}, /* ES1 */
	{0x19F, 0x2F, 0x03, 0x01}  /* ES2 */
};

/* PER DPLL values are same for both ES1 and ES2 */
static struct dpll_param cm_t3530_per_dpll_param[] = {
	/* 12MHz */
	{0x0D8, 0x05, 0x07, 0x09},
	/* 13MHz */
	{0x1B0, 0x0C, 0x03, 0x09},
	/* 19.2MHz */
	{0x0E1, 0x09, 0x07, 0x09},
	/* 26MHz */
	{0x0D8, 0x0C, 0x07, 0x09},
	/* 38.4MHz */
	{0x0E1, 0x13, 0x07, 0x09}
};

/* NAND GPMC config table */
static u32 cm_t3530_nand_gpmc_config[GPMC_MAX_REG] = {
	SMNAND_GPMC_CONFIG1,
	SMNAND_GPMC_CONFIG2,
	SMNAND_GPMC_CONFIG3,
	SMNAND_GPMC_CONFIG4,
	SMNAND_GPMC_CONFIG5,
	SMNAND_GPMC_CONFIG6,
	0
};

static inline int cpu_is_omap34xx(void)
{
	return is_cpu_family() == CPU_OMAP34XX;
}

int cpu_is_omap35xx(void)
{
	return cpu_is_omap34xx();
}

static void cm_t3530_set_muxconf_regs(void)
{
	set_sdrc_muxconf_regs();
	set_gpmc_muxconf_regs();
	set_jtag_muxconf_regs();

	/* CM-T3530 specific mux settings */
	MUX_VAL(CP(GPMC_nCS0),		(IDIS | PTU | EN  | M0));
	MUX_VAL(CP(GPMC_WAIT0),		(IEN  | PTU | EN  | M0));
	MUX_VAL(CP(GPMC_WAIT1),		(IEN  | PTU | EN  | M0));
	MUX_VAL(CP(UART3_TX_IRTX),	(IDIS | PTD | DIS | M0)); /*UART3_TX*/
	MUX_VAL(CP(UART3_RX_IRRX),	(IEN  | PTD | DIS | M0)); /*UART3_RX*/
	MUX_VAL(CP(SYS_32K),		(IEN  | PTD | DIS | M0));
	MUX_VAL(CP(SYS_BOOT0),		(IEN  | PTD | DIS | M4)); /*GPIO2*/
	MUX_VAL(CP(SYS_BOOT1),		(IEN  | PTD | DIS | M4)); /*GPIO3*/
	MUX_VAL(CP(SYS_BOOT2),		(IEN  | PTD | DIS | M4)); /*GPIO4*/
	MUX_VAL(CP(SYS_BOOT3),		(IEN  | PTD | DIS | M4)); /*GPIO5*/
	MUX_VAL(CP(SYS_BOOT4),		(IEN  | PTD | DIS | M4)); /*GPIO6*/
	MUX_VAL(CP(SYS_BOOT5),		(IEN  | PTD | DIS | M4)); /*GPIO7*/
	MUX_VAL(CP(SYS_BOOT6),		(IEN  | PTD | DIS | M4)); /*GPIO8*/
	MUX_VAL(CP(SYS_CLKOUT2),	(IDIS | PTD | DIS | M4)); /*GPIO186*/
	MUX_VAL(CP(sdrc_cke0),		(IDIS | PTU | EN  | M0));
	MUX_VAL(CP(sdrc_cke1),		(IDIS | PTD | DIS | M7)); /*not used*/
}

static void cm_t3530_set_board_data(struct board_data *data)
{
	if (!data)
		hang();

	data->board_name		= "CM-T3530";
	data->watchdog_init		= watchdog_init;
	data->try_unlock_memory		= try_unlock_memory;
	data->set_muxconf_regs		= cm_t3530_set_muxconf_regs;
	data->prcm_init			= prcm_init;
	data->per_clocks_enable		= per_clocks_enable;
	data->setup_memory_controller	= setup_sdrc_memory_controller;
	data->core_dpll_param		= cm_t3530_core_dpll_param;
	data->per_dpll_param		= cm_t3530_per_dpll_param;
	data->mpu_dpll_param		= cm_t3530_mpu_dpll_param;
	data->iva_dpll_param		= cm_t3530_iva_dpll_param;
	data->nand_gpmc_config		= cm_t3530_nand_gpmc_config;
}

int cm_t3530_probe(struct board_data *data)
{
	if (!cpu_is_omap35xx())
		return 1;

	cm_t3530_set_board_data(data);
	return 0;
}

